rm(list = ls())
graphics.off()
library(ncdf)

# =============================================================================
# Author - François Massonnet
#
# Date   - 02 February 2016, Candelaria! 
#          https://es.wikipedia.org/wiki/Fiesta_de_la_Candelaria
#
# Goal   - Create perturbed versions of the fields of DFS5.2. As many
#          perturbations as the user wants can be created, but they will remain
#          in the subspace spanned by the available perturbations (computed in
#          the first script preprocess.bash)
#          A seed is used to ensure reproducibility of perturbations.
#
# Input  - Files of perturbations must be available. They can be computed 
#          using the script preprocess.bash
#
# To do  - Currently the new perturbations retain the spatial and time
#          covariance but not the covariance across variables. This should be
#          done
# =============================================================================

strvar  <- 'u10'               # Name of the variable
fstrvar <- 'u10'               # Name of the variable as it 
                               # appears in the NetCDF
                               # Normally equal to strvar, except 
                               # qsw --> radsw and qlw --> radlw

# The folder where the perturbations are (the folder was created during
# the execution of preprocess.bash)
pertdir <- '/esnas/scratch/fmassonn/TMP/TMP_v10_15558' 

yearbp <- 1958 # The years defining the time period used 
               #to compute perturbations
yearep <- 2015 # That period must match the period for which perturbations were
               # created (in preprocess.bash)

yearb  <- 1958 # The years for which a new perturbation has to be calculated
yeare  <- 2015  

npert <- 29   # Number of perturbations that have to be created each year


# ---------------------------------------------
# Script starts -------------------------------
# ---------------------------------------------

# Assign a seed for variable --> see below. Don't change this, otherwise
# there will be no reproducibility.

switch(strvar,
  t2     = { seedvar = 1},
  q2     = { seedvar = 2},
  u10    = { seedvar = 3},
  v10    = { seedvar = 4},
  precip = { seedvar = 5},
  snow   = { seedvar = 6},
  qsw    = { seedvar = 7},
  qlw    = { seedvar = 8},
  { stop("Seed: variable now known")}
)


nyearp <- yearep - yearbp + 1 # Number of years for the reference period
nyear  <- yeare  - yearb  + 1 # Number of years for which to create perturbations

nsamp <- nyearp - 1          # Number of perturbations samples available. 
                             # Since we create 
                             # the perturbations as successive year-to-year diffs,
                             # we have one less than the number of years

recreate_R <- TRUE           # Set True if the forcing has to be read again,
                             # the SVD decomposition done.


if (recreate_R) {

  # 1/ Creation of the perturbations
  # --------------------------------
  for (year in seq(yearbp + 1, yearep)) {
    print(paste("Reading data:", year, "-", year - 1))
    filein <- paste(pertdir, "/", "diff_", strvar, "_DFS5.2_day_", year, 
                    "-", year - 1, ".nc", sep = "")
    f <- open.ncdf(filein, mode = 'r')
    var <- get.var.ncdf(f, fstrvar)
    if (year == yearbp + 1) {
      # If we are in the first year, let's also create the matrix that will 
      # receive the data
      ny <- dim(var)[1]
      nx <- dim(var)[2]
      nt <- dim(var)[3]

      ns <- ny * nx * nt
  
      lat0 <- get.var.ncdf(f, 'lat0')
      lon0 <- get.var.ncdf(f, 'lon0')
  
      print("Creating X_raw, this has to be done only once")
      X_raw <- matrix(NA, ns, nsamp)
      print("X_raw created")
    }
    close.ncdf(f)
    var[is.na(var)] <- 0
    # This step is necessary, because the DFS5.2 forcing set has, sometimes,
    # NaNs. (See e.g., year 2008, variable t2, time step 1776, in the Southern)
   X_raw[, year - (yearbp + 1) + 1] <- var[]

   rm(list = "var")
  }

  # Centering X
  # -----------
  print("Centering the data")
  myones <- matrix(1, nrow = nsamp, ncol = 1)

  Xbar <- 1 / nsamp * X_raw %*% myones
  X    <- X_raw - Xbar %*% t(myones)

  # Clearing X_raw and Xbar, as they are taking resources!
  rm(list = c("X_raw", "Xbar", "myones"))
  
  print("Data centered")

  # SVD decomposition
  # -----------------
  print("SVD decomposition, this could take a while")
  sv <- svd(X, nu = nsamp, nv = 0)
  print("SVD decomposition done")

  R <- sv$u %*% sqrt(diag(sv$d) %*% t(diag(sv$d))) / sqrt(nsamp - 1)
  rm(list = "sv")

  print("Saving the R matrix")
  save(R, nt, ny, nx, lon0, lat0, file = paste(pertdir, "Rmatrix_", strvar, "_", yearbp, "-", yearep, ".Rdata", sep = ""))
  print("Matrix R saved")
} else {
  print("Loading existing R matrix")
  load(paste(pertdir, "Rmatrix_", strvar, "_", yearbp, "-", yearep, ".Rdata", sep = ""))
  print("Matrix R loaded")
}


# Creating perturbations
# ----------------------
# We take a N(0,1) random vector z of size nsamp by 1
# so that the product R %*% z will give us a new perturbation
# with the desired properties

print("Creating perturbations")

for (year in seq(yearb, yeare)) {
  print(paste("Recording perturbations for year", year))
  
  for (jpert in seq(1, npert)) {
    print(paste("Recording perturbation for member", sprintf("%02d", jpert)))

    # EXTREMELY IMPORTANT - this ensures reproducibility. DON'T CHANGE.
    # if year 1987, variable q2, and perturbation 302:
    # then the seed is 10000 * 1986 + 1000 * 2 + 302 = 19872302. 
    # This is then a unique identifier. It won't be unique if we go beyond 999 
    # members, but I might no longer be on this planet when that happens.
    set.seed(year * 10000 + 1000 * seedvar + jpert)

    z <- matrix(rnorm(nsamp))
    sample <- array(R %*% z, c(ny, nx, nt))
    rm(list = "z")

    # Record in  NetCDF
    # ----------------
    # 1/ Create dimensions
    # --------------------
    dimt <- dim.def.ncdf('time', 'time', seq(1, nt), unlim = TRUE)
    dimy <- dim.def.ncdf('lon0', 'lon0', lon0)
    dimx <- dim.def.ncdf('lat0', 'lat0', lat0)

    # 2/ Define variables
    # -------------------
    mylist <- list()
    mylist[[1]] <- var.def.ncdf(fstrvar, '-', list(dimy, dimx, dimt), -1e24)
    # 3/ Create NetCDF
    # ----------------
    fileout <- paste(pertdir, '/pert_', strvar, '_DFS5.2_', year, '_fc', sprintf("%02d", jpert), '.nc', sep = "")
    fo <- create.ncdf(fileout, mylist)
    close.ncdf(fo)

    # 4/ Open it
    # ----------
    fo <- open.ncdf(fileout, write = TRUE)

    # 5/ Put variable
    # ---------------
    put.var.ncdf(fo, fstrvar, sample)

    # 6/ Close it
    # -----------
    close.ncdf(fo)

    # 7/ Clear workspace
    # ------------------
    rm(list = "sample")
  }
}

